{-# LANGUAGE RecordWildCards #-}

import Affection
import qualified SDL
import qualified GEGL as G
import qualified BABL as B
import qualified Data.Map.Strict as M
import qualified Control.Monad.Parallel as MP

import Debug.Trace

main :: IO ()
main = do
  conf <- return AffectionConfig
    { initComponents = All
    , windowTitle    = "Affection: example00"
    , windowConfig   = SDL.defaultWindow
    , preLoop        = return ()
    , eventLoop      = handle
    , updateLoop     = update
    , drawLoop       = draw
    , loadState      = load
    , cleanUp        = clean
    , canvasSize     = Nothing
    }
  withAffection conf

data UserData = UserData
  { nodeGraph    :: M.Map String G.GeglNode
  , actors       :: M.Map String (Actor String)
  , foreground   :: G.GeglBuffer
  , lastTick     :: Double
  , updateActors :: [Actor String]
  }

load :: IO UserData
load = do
  traceM "loading"
  root <- G.gegl_node_new
  traceM "new root node"
  checkerboard <- G.gegl_node_new_child root $ G.checkerboardOperation $
    props $ do
      prop "color1" $ G.RGBA 0.4 0.4 0.4 1
      prop "color2" $ G.RGBA 0.6 0.6 0.6 1
  traceM "checkerboard"
  over <- G.gegl_node_new_child root G.defaultOverOperation
  traceM "over"
  buffer <- G.gegl_buffer_new (Just $ G.GeglRectangle 0 0 800 600) =<<
    B.babl_format (B.PixelFormat B.RGBA B.CFfloat)
  sink <- G.gegl_node_new_child root $ G.Operation "gegl:copy-buffer" $
    props $
    prop "buffer" buffer
  traceM "buffer-sink"
  rectProps <- return $
    props $ do
      prop "x" (0::Double)
      prop "y" (0::Double)
      prop "width" (20::Double)
      prop "height" (20::Double)
      prop "color" $ G.RGBA 1 0 0 0.5
  rect <- G.gegl_node_new_child root $ G.Operation "gegl:rectangle" rectProps
  traceM "rect"
  crop <- G.gegl_node_new_child root $ G.Operation "gegl:crop" $
    props $ do
      prop "width" (800::Double)
      prop "height" (600::Double)
  G.gegl_node_link_many [checkerboard, over, crop, sink]
  _ <- G.gegl_node_connect_to rect "output" over "aux"
  rectActor <- return $ Actor (map
    (\p -> ActorProperty
      (G.propertyName p)
      (G.propertyValue p)
      (Just "rect")
      )rectProps
    ) (M.singleton "rect" rect)
  traceM "connections made"
  myMap <- return $ M.fromList
    [ ("root"        , root)
    , ("over"        , over)
    , ("background"  , checkerboard)
    , ("sink"        , sink)
    , ("rect"        , rect)
    , ("crop"        , crop)
    ]
  traceM "loading complete"
  actorMap <- return $ M.fromList
    [ ("rect", rectActor)
    ]
  return UserData
    { nodeGraph    = myMap
    , actors       = actorMap
    , foreground   = buffer
    , lastTick     = 0
    , updateActors = []
    }

-- drawInit :: Affection UserData ()
-- drawInit = do
--   UserData{..} <- getAffection
--   present (GeglRectangle 0 0 800 600) foreground True

draw :: Affection UserData ()
draw = do
  ud@UserData{..} <- getAffection
  MP.mapM_ applyProperties updateActors
  putAffection ud
    { updateActors = []
    }
  process (nodeGraph M.! "sink")
  present (GeglRectangle 0 0 800 600) foreground True
  render Nothing Nothing

update :: Double -> Affection UserData ()
update dt = do
  traceM "updating"

  tick <- getElapsedTime
  ud <- getAffection
  putAffection $ ud { lastTick = tick }

  -- let dt = tick - lastTick ud
  return ()
  traceM $ show (1 / dt) ++ " FPS"

handle :: SDL.EventPayload -> Affection UserData ()
handle (SDL.MouseMotionEvent dat) = do
  let (SDL.P (SDL.V2 x y)) = SDL.mouseMotionEventPos dat
  ud <- getAffection
  
  nmap <- return $ M.adjust
    (updateProperties $ props $ do
      prop "x" (fromIntegral (x - 10) :: Double)
      prop "y" (fromIntegral (y - 10) :: Double)
      )
    "rect"
    (actors ud)
  putAffection ud
    { actors       = nmap
    , updateActors = (actors ud M.! "rect") : []
    }
  -- liftIO $ G.gegl_node_set (nodeGraph ud M.! "rect") $ G.Operation "" $
  --   props $ do
  --     prop "x" (fromIntegral (x - 10) :: Double)
  --     prop "y" $ (fromIntegral (y - 10) :: Double)

handle (SDL.WindowClosedEvent _) = do
  traceM "seeya!"
  quit

handle _ =
  return ()

clean :: UserData -> IO ()
clean _ = return ()
